/*
**    TP CPE Lyon
**    Copyright (C) 2014 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#pragma once

#ifndef MATRIX2_HPP
#define MATRIX2_HPP

#include <iostream>
#include <vector>



namespace cpe
{
class vec2;


/** \brief Matrix 2x2 */
class matrix2
{

public:

    // ********************************************* //
    // ********************************************* //
    //  CONSTRUCTORS
    // ********************************************* //
    // ********************************************* //

    /** \brief empty constructor (identity) */
    matrix2();
    /** \brief direct constructor with every value */
    matrix2(float x00,float x01,
            float x10,float x11);

    // ********************************************* //
    // ********************************************* //
    //  Special initialization
    // ********************************************* //
    // ********************************************* //

    /** \brief build identity matrix */
    static matrix2 identity();
    /** \brief build zero matrix */
    static matrix2 zeros();
    /** \brief build rotation matrix */
    static matrix2 rotation(float angle);
    /** \brief build scaling matrix */
    static matrix2 scale(float s);
    /** \brief build scaling matrix */
    static matrix2 scale(float s_x,float s_y);


    // ********************************************* //
    // ********************************************* //
    //  Element access
    // ********************************************* //
    // ********************************************* //

    /** \brief Access to the k_th entry (k in [0,1]) */
    float operator()(const size_t& k1,const size_t& k2) const;
    /** \brief Access to the k_th entry (k in [0,1]) */
    float& operator()(const size_t& k1,const size_t& k2);

    /** \brief fast pointer access
            \warning unprotected access */
    const float* pointer() const;
    /** \brief fast pointer access
            \warning unprotected access */
    float* pointer_unprotected();



    // ********************************************* //
    // ********************************************* //
    //  Math operator
    // ********************************************* //
    // ********************************************* //

    /** \brief + operator */
    matrix2 operator+(const matrix2& m2) const;
    /** \brief + operator */
    matrix2 operator+(float s) const;

    /** \brief - operator */
    matrix2 operator-(const matrix2& m2) const;
    /** \brief - operator */
    matrix2 operator-(float s) const;


    /** \brief multiply by a scalar operator */
    matrix2 operator*(float s) const;
    /** \brief multiply matrix with a v3 */
    vec2 operator*(const vec2& v) const;
    /** \brief matrix mulitplication */
    matrix2 operator*(const matrix2& m2) const;
    /** \brief divide by a scalar operator */
    matrix2 operator/(float s) const;


    /** \brief internal + */
    matrix2& operator+=(const matrix2& m);
    /** \brief internal + */
    matrix2& operator+=(float s);
    /** \brief internal - */
    matrix2& operator-=(const matrix2& m);
    /** \brief internal - */
    matrix2& operator-=(float s);
    /** \brief internal * */
    matrix2& operator*=(float s);
    /** \brief internal / */
    matrix2& operator/=(float s);

    /** \brief unary negation */
    matrix2 operator-() const;

    /** \brief does componentwise multiplication */
    matrix2 product_compontentwise(const matrix2& m) const;
    /** \brief does componentwise multiplication */
    matrix2& product_compontentwise_internal(const matrix2& m);

    /** \brief transpose matrix */
    matrix2 transposed() const;







    // ********************************************* //
    // ********************************************* //
    //  Projection matrix
    // ********************************************* //
    // ********************************************* //

    /** \brief convert 3x3 matrix into a vector of float of size 9 */
    std::vector<float> to_vector() const;

private:

    // ********************************************* //
    // ********************************************* //
    //  Internal parameters
    // ********************************************* //
    // ********************************************* //

    /** \brief internal storage of the matrix */
    float m[4];


    // ********************************************* //
    // ********************************************* //
    //  Helper
    // ********************************************* //
    // ********************************************* //

    /** \brief assert that a size_t belongs to [[0,1]] */
    void assert_size(const size_t& k1,const size_t& k2) const;

};

/** \brief + operator */
matrix2 operator+(float s,const matrix2& m);
/** \brief - operator */
matrix2 operator-(float s,const matrix2& m);
/** \brief multiply by a scalar operator */
matrix2 operator*(float s,const matrix2& m);

// ********************************************* //
// ********************************************* //
//  Output
// ********************************************* //
// ********************************************* //

/** \brief output the vector in ostream */
std::ostream& operator<<(std::ostream& stream,const matrix2& m);






}

#endif
