/*
**    TP CPE Lyon
**    Copyright (C) 2014 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#pragma once

#ifndef MATRIX4_HPP
#define MATRIX4_HPP

#include <iostream>
#include <vector>



namespace cpe
{
class vec3;
class vec4;
class matrix3;

/** \brief Matrix 4x4 */
class matrix4
{


public:

    // ********************************************* //
    // ********************************************* //
    //  CONSTRUCTORS
    // ********************************************* //
    // ********************************************* //

    /** \brief empty constructor (identity) */
    matrix4();
    /** \brief direct constructor with every value */
    matrix4(float x00,float x01,float x02,float x03,
            float x10,float x11,float x12,float x13,
            float x20,float x21,float x22,float x23,
            float x30,float x31,float x32,float x33);
    /** \brief direct constructor from matrix 3 */
    matrix4(const matrix3& m);

    // ********************************************* //
    // ********************************************* //
    //  Special initialization
    // ********************************************* //
    // ********************************************* //

    /** \brief build identity matrix */
    static matrix4 identity();
    /** \brief build zero matrix */
    static matrix4 zeros();
    /** \brief build rotation matrix */
    static matrix4 rotation(const vec3& axis,float angle);
    /** \brief build scaling matrix */
    static matrix4 scale(float s);
    /** \brief build scaling matrix */
    static matrix4 scale(float s_x,float s_y,float s_z,float s_w);
    /** \brief build translation matrix */
    static matrix4 translation(const vec3& translation);
    /** \brief build from matrix3 and translation */
    static matrix4 transformation(const matrix3& m3,const vec3& translation);

    // ********************************************* //
    // ********************************************* //
    //  Projection matrix (emulate glu(t) functions)
    // ********************************************* //
    // ********************************************* //

    /** \brief build a perspective projection similar to gluPerspective
        \param float fovy: in radian
         see: http://www.opengl.org/sdk/docs/man/xhtml/gluPerspective.xml
    */
    static matrix4 projection_perspective(float fovy,float aspect,float z_near,float z_far);
    /** \brief build a perspective projection (frustum) similar to gluPerspective
         see: http://www.opengl.org/sdk/docs/man/xhtml/glFrustum.xml
    */
    static matrix4 projection_frustum(float left,float right,float bottom,float top,float near,float far);
    /** \brief build an orthographic projection similar to gluPerspective
         see: http://www.opengl.org/sdk/docs/man/xhtml/glOrtho.xml
    */
    static matrix4 projection_orthographic(float left,float right,float bottom,float top,float near,float far);
    /** \brief build a perspective projection similar to gluLookAt
         see: http://www.opengl.org/sdk/docs/man/xhtml/gluLookAt.xml
    */
    static matrix4 projection_look_at(const matrix4& current_matrix,const vec3& eye,const vec3& center,const vec3& up);


    // ********************************************* //
    // ********************************************* //
    //  Element access
    // ********************************************* //
    // ********************************************* //

    /** \brief Access to the k_th entry (k in [0,3]) */
    float operator()(const size_t& k1,const size_t& k2) const;
    /** \brief Access to the k_th entry (k in [0,3]) */
    float& operator()(const size_t& k1,const size_t& k2);

    /** \brief fast pointer access
            \warning unprotected access */
    const float* pointer() const;
    /** \brief fast pointer access
            \warning unprotected access */
    float* pointer_unprotected();



    // ********************************************* //
    // ********************************************* //
    //  Math operator
    // ********************************************* //
    // ********************************************* //

    /** \brief + operator */
    matrix4 operator+(const matrix4& m2) const;
    /** \brief + operator */
    matrix4 operator+(float s) const;

    /** \brief - operator */
    matrix4 operator-(const matrix4& m2) const;
    /** \brief - operator */
    matrix4 operator-(float s) const;


    /** \brief multiply by a scalar operator */
    matrix4 operator*(float s) const;
    /** \brief multiply matrix with a v4 */
    vec4 operator*(const vec4& v) const;
    /** \brief multiply matrix with a v3 (assume (v3.x,v3.y,v3.z,1) and normalize)*/
    vec3 operator*(const vec3& v) const;
    /** \brief matrix mulitplication */
    matrix4 operator*(const matrix4& m2) const;
    /** \brief divide by a scalar operator */
    matrix4 operator/(float s) const;


    /** \brief internal + */
    matrix4& operator+=(const matrix4& m);
    /** \brief internal + */
    matrix4& operator+=(float s);
    /** \brief internal - */
    matrix4& operator-=(const matrix4& m);
    /** \brief internal - */
    matrix4& operator-=(float s);
    /** \brief internal * */
    matrix4& operator*=(float s);
    /** \brief internal / */
    matrix4& operator/=(float s);

    /** \brief unary negation */
    matrix4 operator-() const;

    /** \brief does componentwise multiplication */
    matrix4 product_compontentwise(const matrix4& m) const;
    /** \brief does componentwise multiplication */
    matrix4& product_compontentwise_internal(const matrix4& m);

    /** \brief transpose matrix */
    matrix4 transposed() const;

    // ********************************************* //
    // ********************************************* //
    //  Deformation
    // ********************************************* //
    // ********************************************* //

    /** \brief Translated matrix (return of copy of the translated matrix) */
    matrix4 translated(const vec3& translation) const;
    /** \brief Translated matrix (modify the matrix) */
    void translate_internal(const vec3& translation);






    // ********************************************* //
    // ********************************************* //
    //  Projection matrix
    // ********************************************* //
    // ********************************************* //

    /** \brief convert 4x4 matrix into a 16 vector of float (glMultmatrixf) */
    std::vector<float> to_vector() const;

private:

    // ********************************************* //
    // ********************************************* //
    //  Internal parameters
    // ********************************************* //
    // ********************************************* //

    /** \brief internal storage of the matrix */
    float m[16];


    // ********************************************* //
    // ********************************************* //
    //  Helper
    // ********************************************* //
    // ********************************************* //

    /** \brief assert that a size_t belongs to [[0,3]] */
    void assert_size(const size_t& k1,const size_t& k2) const;

};

// ********************************************* //
// ********************************************* //
//  Output
// ********************************************* //
// ********************************************* //

/** \brief output the vector in ostream */
std::ostream& operator<<(std::ostream& stream,const matrix4& m);




/** \brief + operator */
matrix4 operator+(float s,const matrix4& m);
/** \brief - operator */
matrix4 operator-(float s,const matrix4& m);
/** \brief multiply by a scalar operator */
matrix4 operator*(float s,const matrix4& m);

}

#endif
