/*
**    TP CPE Lyon
**    Copyright (C) 2014 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include "navigator_tool.hpp"

#include <cmath>

#include "../lib3d/matrix3.hpp"

namespace cpe
{

    navigator_tool::navigator_tool()
        :x_size(800),y_size(800),mouse_x_previous(0),mouse_y_previous(0),is_left_button(false),is_right_button(false)
    {
        fov=45.0*M_PI/180.0;
        znear=1e-2;
        zfar=150;
        dist=-5.0;

    }

    unsigned int& navigator_tool::screen_size_x(){return x_size;}
    unsigned int navigator_tool::screen_size_x() const{return x_size;}

    unsigned int& navigator_tool::screen_size_y(){return y_size;}
    unsigned int navigator_tool::screen_size_y() const{return y_size;}

    double& navigator_tool::dist_cam() {return dist;}
    float navigator_tool::dist_cam() const {return dist;}

    void navigator_tool::go_forward(float d_L)
    {
        vec3 z(0,0,1);
        tr += d_L*(track.quat().rotate(z));
    }

    void navigator_tool::go_right  (float d_L)
    {
        vec3 x(-1,0,0);
        tr += d_L*(track.quat().rotate(x));
    }

    void navigator_tool::go_up     (float d_L)
    {
        vec3 y(0,1,0);
        tr += d_L*(track.quat().rotate(y));
    }

    int& navigator_tool::x_previous() {return mouse_x_previous;}
    const int& navigator_tool::x_previous() const {return mouse_x_previous;}

    int& navigator_tool::y_previous() {return mouse_y_previous;}
    const int& navigator_tool::y_previous() const {return mouse_y_previous;}


    bool& navigator_tool::left_button() {return is_left_button;}
    bool navigator_tool::left_button() const {return is_left_button;}

    bool& navigator_tool::right_button() {return is_right_button;}
    bool navigator_tool::right_button() const {return is_right_button;}

    void navigator_tool::trackball_rotate(const int& x,const int&y)
    {
        float x_old=mouse_x_previous;
        float y_old=mouse_y_previous;
        float h=y_size;
        float w=x_size;

        double x0=(2.0*x_old-w)/w;
        double y0=(2.0*y_old-h)/h;
        double x1=(2*x-w)/w;
        double y1=(2*y-h)/h;
        if(std::sqrt((x0-x1)*(x0-x1)+(y0-y1)*(y0-y1))>1e-4)
        {
            track.set_2d_coords(x0,y0,x1,y1);
            track.apply_rotation();
        }
        mouse_x_previous=x;
        mouse_y_previous=y;
    }

    void navigator_tool::trackball_zoom(const int& y)
    {
        //magnification factor
        const double F0=500;

        float y_old=mouse_y_previous;

        double u = y-y_old;
        double fu= u/F0;

        dist += (std::fabs(dist)+1)*fu;
        dist = std::min(dist,0.0);
        mouse_y_previous=y;

    }

    std::vector<double> navigator_tool::projection_param() const
    {
        double aspect=static_cast<double>(x_size)/static_cast<double>(y_size);
        double top=std::tan(fov*0.5)*znear;
        double left=-aspect*top;
        std::vector<double> param;
        param.push_back(left);
        param.push_back(top);
        param.push_back(znear);
        param.push_back(zfar);
        param.push_back(aspect);

        return param;
    }

    matrix3 navigator_tool::camera_orientation() const
    {
        return track.quat().conjugated().matrix();
    }
    vec3 navigator_tool::camera_translation() const
    {
        return tr;
    }
    vec3 navigator_tool::camera_position() const
    {
        vec3 p(0,0,dist_cam());
        p=camera_orientation().transposed()*p;

        p+=tr;
        return -p;
    }

    void navigator_tool::increase_frame_number() {++frame;}

    unsigned int navigator_tool::time() const {return current_time;}
    unsigned int& navigator_tool::time() {return current_time;}

    double navigator_tool::update_fps()
    {
        current_fps=static_cast<double>(frame*1000)/static_cast<double>(current_time-old_time);
        old_time=current_time;
        frame=0;
        return current_fps;
    }
    double navigator_tool::delta_time()
    {
        return current_time-old_time;
    }

    double navigator_tool::fps()
    {
        return current_fps;
    }

    std::pair<vec3,vec3> navigator_tool::ray_world_space_cam1(const int& x_screen,const int& y_screen) const
    {
        //normalized screen coordinates
        vec3 local(
                    (static_cast<int>(screen_size_x())-2*x_screen)/static_cast<double>(screen_size_x())
                    ,
                    (static_cast<int>(screen_size_y())-2*y_screen)/static_cast<double>(screen_size_y()),0);

        std::vector<double> proj_param=projection_param();
        double Lh=-proj_param[1];
        double aspect=proj_param[4];

        vec3 proj_ray(local[0]*Lh*aspect,local[1]*Lh,-znear);

        matrix3 mod=camera_orientation();
        matrix3 mod_inv=mod.transposed();

        vec3 ray=(mod_inv*proj_ray).normalized();
        vec3 center=camera_position();

        return std::make_pair(center,ray);
    }


}
